@@ -1,30 +1,41 @@
-# -*- coding: utf-8 -*-
+from __future__ import annotations
+
 import math
 import warnings
-from collections.abc import Iterator
 from collections import deque
 from itertools import (
-    zip_longest,
-    islice,
-    starmap,
-    takewhile,
-    tee,
     chain,
+    combinations,
+    combinations_with_replacement,
     compress,
+    count,
+    cycle,
     dropwhile,
     filterfalse,
     groupby,
-    repeat,
-    product,
-    combinations,
-    combinations_with_replacement,
+    islice,
     permutations,
-    cycle,
-    count,
+    product,
+    repeat,
+    starmap,
+    takewhile,
+    tee,
+    zip_longest,
+)
+from typing import (
+    Any,
+    Callable,
+    Deque,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Tuple,
+    TypeVar,
+    Union,
 )
-from typing import Iterable, Optional, Callable, Any, List, Iterator as IteratorType
 
-from f_it.utils import len_or_none, nCr, n_permutations
+from f_it.utils import len_or_none, n_permutations, nCr
 
 
 def ensure_FIt(obj):
@@ -34,8 +45,15 @@ def ensure_FIt(obj):
         return FIt(obj)
 
 
-class FIt(Iterator):
-    def __init__(self, iterable: Iterable, length=None):
+neg_idx_msg = "Negative indices into FIt are only possible when it has a known length"
+
+EMPTY = object()
+
+T = TypeVar("T")
+
+
+class FIt(Iterator[T]):
+    def __init__(self, iterable: Iterable[T], length=None):
         """Iterator class providing many postfix functional methods.
 
         Most of these methods can also be used as static methods which take any
@@ -55,7 +73,7 @@ class FIt(Iterator):
             self.init_length = int(length)
 
         self.consumed = 0
-        self.cache = deque()
+        self.cache: Deque[T] = deque()
 
     @property
     def length(self) -> Optional[int]:
@@ -75,7 +93,7 @@ class FIt(Iterator):
         else:
             return length
 
-    def __next__(self):
+    def __next__(self) -> T:
         if self.cache:
             return self.cache.popleft()
         else:
@@ -83,19 +101,51 @@ class FIt(Iterator):
             self.consumed += 1
             return item
 
-    def __add__(self, other: IteratorType):
-        if isinstance(other, Iterator):
+    def __add__(self, other: Iterator[T]):
+        """Shorthand for ``FIt.chain(self, other)``.
+
+        Requires ``other`` to be an ``Iterator`` (not just an ``Iterable``).
+        """
+        try:
             return self.chain(other)
+        except Exception:
+            pass
         return NotImplemented
 
-    def __radd__(self, other: IteratorType):
+    def __radd__(self, other: Iterator[T]):
+        """Shorthand for ``FIt.chain(other, self)``.
+
+        Requires ``other`` to be an ``Iterator`` (not just an ``Iterable``!).
+        """
         if isinstance(other, FIt):
             return other.chain(self)
-        elif isinstance(other, Iterator):
-            return FIt(other).chain(self)
+        else:
+            try:
+                return FIt(other).chain(self)
+            except TypeError:
+                pass
         return NotImplemented
 
-    def next(self):
+    def __getitem__(self, idx: Union[int, slice]):
+        """Shorthand for ``.get`` or ``.islice``.
+
+        Depending on whether an integer or slice is given,
+        returns either a single item (see ``.get``),
+        or an ``FIt`` instance (see ``.islice``).
+        Consumes necessary iterator elements.
+        """
+        if isinstance(idx, int):
+            if idx < 0 and len_or_none(self) is None:
+                raise ValueError(neg_idx_msg)
+            return self.get(idx)
+        elif isinstance(idx, slice):
+            return self.islice(idx.start, idx.stop, idx.step)
+        else:
+            raise TypeError(
+                "FIt indices must be integers or slices, not " + type(idx).__name__
+            )
+
+    def next(self) -> T:
         """Return the next item in the iterator"""
         return next(self)
 
@@ -115,21 +165,21 @@ class FIt(Iterator):
     # Free functions #
     ##################
 
-    def enumerate(self, start=0):
+    def enumerate(self, start=0) -> FIt[Tuple[int, T]]:
         """See the builtin enumerate_
 
         .. _enumerate: https://docs.python.org/3/library/functions.html#enumerate
         """
         return FIt(enumerate(self, start), len_or_none(self))
 
-    def filter(self, function=None):
+    def filter(self, function=None) -> FIt[T]:
         """See the builtin filter_
 
         .. _filter: https://docs.python.org/3/library/functions.html#filter
         """
         return FIt(filter(function, self))
 
-    def map(self, function, *iterables, longest=True):
+    def map(self, function, *iterables, longest=True) -> FIt:
         """See the builtin map_
 
         Difference from stdlib: accepts ``longest`` kwarg, for whether to terminate
@@ -141,7 +191,7 @@ class FIt(Iterator):
         return FIt.zip(self, *iterables, longest=longest).starmap(function)
 
     @classmethod
-    def range(cls, start, stop=None, step=1):
+    def range(cls, start, stop=None, step=1) -> FIt[int]:
         """See the builtin range_
 
         .. _range: https://docs.python.org/3/library/functions.html#func-range
@@ -152,11 +202,16 @@ class FIt(Iterator):
 
         return FIt(range(start, stop, step))
 
-    def zip(self, *iterables, longest=False, fill_value=None):
+    def zip(
+        self, *iterables, longest=False, fill_value=None, strict=False
+    ) -> FIt[Tuple]:
         """See the builtin zip_
 
         Difference from stdlib: accepts ``longest`` kwarg, which makes this method
-        act like itertools.zip_longest_ (also accepts ``fill_value``)
+        act like itertools.zip_longest_ (also accepts ``fill_value``).
+
+        ``strict=True`` is only supported in python >=3.10
+        and will raise a TypeError in lower versions.
 
         .. _zip: https://docs.python.org/3/library/functions.html#zip
         .. _zip_longest: https://docs.python.org/3/library/itertools.html#itertools.zip_longest
@@ -171,7 +226,12 @@ class FIt(Iterator):
         except TypeError:
             length = None
 
-        return FIt(zip(*iterables), length)
+        if strict:
+            inner = zip(*iterables, strict=True)
+        else:
+            inner = zip(*iterables)
+
+        return FIt(inner, length)
 
     #############
     # functools #
@@ -198,7 +258,7 @@ class FIt(Iterator):
     # itertools #
     #############
 
-    def chain(self, *iterables):
+    def chain(self, *iterables: Iterable[T]) -> FIt[T]:
         """See itertools.chain_
 
         .. _itertools.chain: https://docs.python.org/3/library/itertools.html#itertools.chain
@@ -212,7 +272,7 @@ class FIt(Iterator):
 
         return FIt(chain.from_iterable(iterables), length)
 
-    def chain_from_iterable(self, iterable):
+    def chain_from_iterable(self, iterable: Iterable[Iterable[T]]) -> FIt[T]:
         """See itertools.chain.from_iterable_
 
         .. _itertools.chain.from_iterable: https://docs.python.org/3/library/itertools.html#itertools.chain.from_iterable
@@ -225,7 +285,7 @@ class FIt(Iterator):
 
         return FIt(gen())
 
-    def combinations(self, r: int, replace=False):
+    def combinations(self, r: int, replace=False) -> FIt[Tuple[T, ...]]:
         """See itertools.combinations_
 
         Difference from stdlib: ``replace`` argument to use
@@ -245,7 +305,7 @@ class FIt(Iterator):
 
         return FIt(combinations(self, r), length)
 
-    def combinations_with_replacement(self, r: int):
+    def combinations_with_replacement(self, r: int) -> FIt[Tuple[T, ...]]:
         """See itertools.combinations_with_replacement_
 
         .. _itertools.combinations_with_replacement: https://docs.python.org/3/library/itertools.html#itertools.combinations_with_replacement
@@ -258,7 +318,7 @@ class FIt(Iterator):
 
         return FIt(combinations_with_replacement(self, r), length)
 
-    def compress(self, selectors: Iterable):
+    def compress(self, selectors: Iterable) -> FIt[T]:
         """See itertools.compress_
 
         .. _itertools.compress: https://docs.python.org/3/library/itertools.html#itertools.compress
@@ -266,14 +326,14 @@ class FIt(Iterator):
         return FIt(compress(self, selectors))
 
     @classmethod
-    def count(cls, start: int = 0, step: int = 1):
+    def count(cls, start: int = 0, step: int = 1) -> FIt[int]:
         """See itertools.count_
 
         .. _itertools.count: https://docs.python.org/3/library/itertools.html#itertools.count
         """  # noqa
         return FIt(count(start, step))
 
-    def cycle(self, n: Optional[int] = None):
+    def cycle(self, n: Optional[int] = None) -> FIt[T]:
         """See itertools.cycle_
 
         Difference from stdlib: accepts ``n`` argument, for how many times it should be
@@ -301,7 +361,7 @@ class FIt(Iterator):
 
         return FIt(gen(self))
 
-    def dropwhile(self, predicate: Optional[Callable[[Any], bool]] = None):
+    def dropwhile(self, predicate: Optional[Callable[[Any], bool]] = None) -> FIt[T]:
         """See itertools.dropwhile_
 
         Difference from stdlib: if predicate is None (default), use ``bool``
@@ -311,7 +371,7 @@ class FIt(Iterator):
         predicate = predicate or bool
         return FIt(dropwhile(predicate, self))
 
-    def filterfalse(self, predicate: Callable[[Any], bool] = None):
+    def filterfalse(self, predicate: Callable[[Any], bool] = None) -> FIt[T]:
         """See itertools.filterfalse_
 
         Difference from stdlib: if predicate is None (default), use ``bool``
@@ -328,24 +388,41 @@ class FIt(Iterator):
         """  # noqa
         return FIt(groupby(self, key))
 
-    def islice(self, start: int, stop: Optional[int] = None, step: int = 1):
+    def islice(
+        self, start: int, stop: Optional[int] = None, step: Optional[int] = None
+    ) -> FIt[T]:
         """See itertools.islice_
 
+        Difference from stdlib: if the FIt length is known, negative indices are allowed,
+        although a negative step size is still not.
+        Consumes elements up to and including the start index;
+        subsequent items are consumed by the returned iterator.
+
         .. _itertools.islice: https://docs.python.org/3/library/itertools.html#itertools.islice
         """  # noqa
         if stop is None:
             stop = start
             start = 0
+        if step is None:
+            step = 1
 
         this_len = len_or_none(self)
         if this_len is None:
+            if start < 0 or stop < 0:
+                raise ValueError(neg_idx_msg)
+
             length = None
         else:
+            if start < 0:
+                start = max(this_len + start, 0)
+            if stop < 0:
+                stop = max(this_len + stop, 0)
+
             length = max(math.ceil((min(stop, this_len) - start) / step), 0)
 
         return FIt(islice(self, start, stop, step), length)
 
-    def permutations(self, r: Optional[int] = None):
+    def permutations(self, r: Optional[int] = None) -> FIt[Tuple[T, ...]]:
         """See itertools.permutations_
 
         .. _itertools.permutations: https://docs.python.org/3/library/itertools.html#itertools.permutations
@@ -358,7 +435,7 @@ class FIt(Iterator):
 
         return FIt(permutations(self, r), length)
 
-    def product(self, *iterables, repeat: int = 1):
+    def product(self, *iterables, repeat: int = 1) -> FIt[Tuple[T, ...]]:
         """See itertools.product_
 
         .. _itertools.product: https://docs.python.org/3/library/itertools.html#itertools.product
@@ -374,7 +451,7 @@ class FIt(Iterator):
         return FIt(product(*iterables, repeat=repeat), length)
 
     @classmethod
-    def repeat(cls, obj, n: Optional[int] = None):
+    def repeat(cls, obj, n: Optional[int] = None) -> FIt:
         """See itertools.repeat_
 
         .. _itertools.repeat: https://docs.python.org/3/library/itertools.html#itertools.repeat
@@ -382,14 +459,14 @@ class FIt(Iterator):
         it = repeat(obj) if n is None else repeat(obj, n)
         return cls(it, n)
 
-    def starmap(self, function: Callable):
+    def starmap(self, function: Callable) -> FIt:
         """See itertools.starmap_
 
         .. _itertools.starmap: https://docs.python.org/3/library/itertools.html#itertools.starmap
         """  # noqa
         return FIt(starmap(function, self), len_or_none(self))
 
-    def takewhile(self, predicate: Optional[Callable[[Any], bool]] = None):
+    def takewhile(self, predicate: Optional[Callable[[Any], bool]] = None) -> FIt[T]:
         """See itertools.takewhile_
 
         .. _itertools.takewhile: https://docs.python.org/3/library/itertools.html#itertools.takewhile
@@ -398,7 +475,7 @@ class FIt(Iterator):
             predicate = bool
         return FIt(takewhile(predicate, self))
 
-    def tee(self, n: int = 2):
+    def tee(self, n: int = 2) -> List[FIt[T]]:
         """See itertools.tee_
 
         .. _itertools.tee: https://docs.python.org/3/library/itertools.html#itertools.tee
@@ -406,7 +483,7 @@ class FIt(Iterator):
         length = len_or_none(self)
         return [FIt(t, length) for t in tee(self, n)]
 
-    def zip_longest(self, *iterables, fill_value=None):
+    def zip_longest(self, *iterables, fill_value=None) -> FIt[Tuple]:
         """See itertools.zip_longest_
 
         .. _itertools.zip_longest: https://docs.python.org/3/library/itertools.html#itertools.zip_longest
@@ -424,14 +501,14 @@ class FIt(Iterator):
     # itertools recipes #
     #####################
 
-    def take(self, n: int) -> List:
+    def take(self, n: int) -> List[T]:
         """Return the first ``n`` items of the iterable as a list
 
         Cannot be safely used as a static method.
         """
         return list(self.islice(n))
 
-    def tail(self, n: int):
+    def tail(self, n: int) -> FIt[T]:
         """Return an iterator over the last ``n`` items"""
         this_len = len_or_none(self)
         if this_len is None:
@@ -448,17 +525,41 @@ class FIt(Iterator):
             next(self.islice(n, n), None)
         return self
 
-    def get(self, n: int, default=None):
-        """Alias for ``FIt.nth``: returns the nth item or a default value
+    def get(self, n: int, default=EMPTY) -> T:
+        """Alias for ``FIt.nth``: returns the nth item or a default value.
+
+        If default is not given, raises IndexError.
+        Accepts negative index if length is known.
+        Consumes elements up to and including the given index.
 
         Cannot be safely used as a static method.
         """
-        return next(self.islice(n, n + 1), default)
+        if n < 0:
+            this_len = len_or_none(self)
+            if this_len is None:
+                raise ValueError(neg_idx_msg)
+            n = this_len + n
+            if n < 0:
+                raise IndexError("FIt index out of range")
+        try:
+            result = next(self.islice(n, n + 1))
+        except StopIteration:
+            if default is EMPTY:
+                raise IndexError("FIt index out of range")
+            else:
+                result = default
+        return result
 
-    def nth(self, n: int, default=None):
-        """Returns the nth item or a default value
+    def nth(self, n: int, default=EMPTY) -> T:
+        """Returns the nth item or a default value.
+
+        If default is not given, raises IndexError.
+        Accepts negative index if length is known.
+        Consumes elements up to and including the given index.
 
         Cannot be safely used as a static method.
+
+        Alias for ``.get()``.
         """
         return self.get(n, default)
 
@@ -504,7 +605,7 @@ class FIt(Iterator):
     # others #
     ##########
 
-    def sliding_window(self, n: int):
+    def sliding_window(self, n: int) -> FIt[Tuple[T, ...]]:
         """Iterate over ``n``-length tuples forming a sliding window over the iterable.
 
         :param n: window size
@@ -524,7 +625,7 @@ class FIt(Iterator):
 
         return FIt(gen(self), length)
 
-    def chunk(self, chunksize: int):
+    def chunk(self, chunksize: int) -> FIt[List[T]]:
         """Iterate over ``chunksize``-or-shorter lists which are chunks of the iterable.
 
         :param chunksize: maximum length for each chunk (all but the last chunk will be this size)
@@ -548,7 +649,7 @@ class FIt(Iterator):
 
         return FIt(gen(self), length)
 
-    def interleave(self, *iterables):
+    def interleave(self, *iterables: Iterable[T]) -> FIt[T]:
         """Interleave items from any number of iterables
 
         When an iterable is exhausted, items continue to be yielded from the remaining
@@ -578,7 +679,7 @@ class FIt(Iterator):
 
         return FIt(gen(iterables), length)
 
-    def peek(self, n: Optional[int] = None):
+    def peek(self, n: Optional[int] = None) -> Union[T, List[T]]:
         """Return a list of the next ``n`` items without advancing the iterator.
 
         If ``n`` is None, return a single item.
@@ -603,7 +704,7 @@ class FIt(Iterator):
 
         return peeked[0] if n is None else peeked
 
-    def for_each(self, function: Callable, *args, **kwargs):
+    def for_each(self, function: Callable, *args, **kwargs) -> None:
         """Consume the iterator, applying a callable on each item.
         Return values are discarded.
 
@@ -615,11 +716,39 @@ class FIt(Iterator):
         for item in self:
             function(item, *args, **kwargs)
 
+    ###########
+    # logical #
+    ###########
+
+    def any(self) -> bool:
+        """True if any item is truthy, False otherwise.
+
+        Consumes as few items as possible.
+
+        :return: bool
+        """
+        for item in self:
+            if item:
+                return True
+        return False
+
+    def all(self) -> bool:
+        """False if any item is falsey, True otherwise.
+
+        Consumes as few items as possible.
+
+        :return: bool
+        """
+        for item in self:
+            if not item:
+                return False
+        return True
+
     ############
     # optional #
     ############
 
-    def progress(self, **kwargs):
+    def progress(self, **kwargs) -> FIt[T]:
         """Create a tqdm progress bar for this iterable.
 
         :param kwargs: passed to tqdm instance
